import numpy as np
import json

from .. import util

def load_off(file_obj, file_type=None):
    '''
    Load an OFF file into the kwargs for a Trimesh constructor


    Arguments
    ----------
    file_obj: file object containing an OFF file
    file_type: not used
    
    Returns
    ----------
    loaded: dict with kwargs for Trimesh constructor (vertices, faces)
 
    '''
    header_string = file_obj.readline().decode().strip()
    if not header_string == 'OFF': 
        raise NameError('Not an OFF file! Header was ' + header_string)

    header  = np.array(file_obj.readline().split()).astype(int)
    blob    = np.array(file_obj.read().split())
    data_ok = np.sum(header * [3,4,0]) == len(blob)    
    if not data_ok: raise NameError('Incorrect number of vertices or faces!')

    vertices = blob[0:(header[0]*3)].astype(float).reshape((-1,3))
    faces    = blob[(header[0]*3):].astype(int).reshape((-1,4))[:,1:]
  
    return {'vertices' : vertices,
            'faces'    : faces}

def load_wavefront(file_obj, file_type=None):
    '''
    Loads an ascii Wavefront OBJ file_obj into kwargs for the Trimesh constructor.

    Discards texture normals and vertex color information.


    Arguments
    ----------
    file_obj: file object containing a wavefront file
    file_type: not used
    
    Returns
    ----------
    loaded: dict with kwargs for Trimesh constructor (vertices, faces)
    '''
    data     = np.array(file_obj.read().split())
    data_str = data.astype(str)
    
    # find the locations of keys, then find the proceeding values
    vid = np.nonzero(data_str == 'v')[0].reshape((-1,1))  + np.arange(3) + 1
    nid = np.nonzero(data_str == 'vn')[0].reshape((-1,1)) + np.arange(3) + 1
    fid = np.nonzero(data_str == 'f')[0].reshape((-1,1))  + np.arange(3) + 1
    
    # if we wanted to use the texture/vertex normals, we could slice this differently.
    faces = np.array([i.split(b'/') for i in data[fid].reshape(-1)])[:,0].reshape((-1,3))
    # wavefront has 1- indexed faces, as opposed to 0- indexed
    faces = faces.astype(int) - 1
    loaded = {'vertices'       : data[vid].astype(float),
              'vertex_normals' : data[nid].astype(float),
              'faces'          : faces}
    return loaded

def load_msgpack(blob, file_type=None):
    import msgpack
    if util.is_file(blob):
        data = msgpack.load(blob)
    else:
        data = msgpack.loads(blob)
    loaded = load_dict(data)
    return loaded

def load_dict(data, file_type=None):
    '''
    Load multiple input types into kwargs for a Trimesh constructor.

    Arguments
    ----------
    data: accepts multiple forms
          -dict: has keys for vertices and faces as (n,3) numpy arrays
          -dict: has keys for vertices/faces (n,3) arrays encoded in base64
          -str:  json blob as dict with either straight array or base64 values
          -file object: json blob of dict
    file_type: not used

    Returns
    -----------
    loaded: dict with keys 
            -vertices: (n,3) float
            -faces:    (n,3) int
            -face_normals: (n,3) float (optional)
    '''
    if util.is_string(data):
        if not '{' in data:
            raise ValueError('Object is not a JSON encoded dictionary!')
        data = json.loads(data.decode('utf-8'))
    elif util.is_file(data):
        data = json.load(data)

    # now go through data structure and if anything is encoded as base64
    # pull it back into numpy arrays
    if util.is_dict(data):
        loaded = {}
        for key in ('vertices', 
                    'faces', 
                    'face_normals'):
            if key in data:
                loaded[key] = util.encoded_to_array(data[key])
        return loaded
    else:
        raise ValueError('%s object passed to dict loader!',
                         data.__class__.__name__)

_misc_loaders = {'obj'     : load_wavefront,
                 'off'     : load_off,
                 'dict'    : load_dict,
                 'dict64'  : load_dict,
                 'json'    : load_dict,
                 'msgpack' : load_msgpack}
